local AceOO = AceLibrary("AceOO-2.0")
local L = AceLibrary("AceLocale-2.2"):new("ag_UnitFrames")
local SurfaceLib = AceLibrary("Surface-1.0")

local print = function(msg) if msg then DEFAULT_CHAT_FRAME:AddMessage(msg) end end

-- This file contains the class and subclasses that controls the unit frames of ag_UnitFrames.
-- It is divided up in the following sections:
--
-- - System
--		- Creates the object, registers events
--
-- - Frame Setup
--		- Sets up the frame, controls theme, size and scale
--
-- - Updates
--		- Contains the UpdateAll method which updates everything when needed, and serveral other
--			updates like strings, portrait, etc
--
-- - Statusbar
--		- Updates the status bars
--
-- - Icons
--		- Controls various status icons, like leader, loot master, pvp, resting, combat
--
-- - Auras
-- 		- Controls the aura positions and display
--
-- - Combat Events
--		- Displays the combat text when units take damage or receive heals
--
--
-- Furthermore the following subclasses exist:
--
-- - XP Subclasses
-- 		- For the player and pet units, which both have an experience bar
--
-- - OnUpdate Subclass
--		- For main tank frames, target of target and unittarget frames
--
-- - Combopoint Subclass
--		- For the target frame

-- - Vanilla Subclass
--		- For all other types of units

------------
-- SYSTEM --
------------

aUF.classes = {}
aUF.classes.aUFunit = AceOO.Class("AceEvent-2.0","AceHook-2.1")
aUF.classes.aUFunitPlayer = AceOO.Class(aUF.classes.aUFunit)
aUF.classes.aUFunitPet = AceOO.Class(aUF.classes.aUFunit)
aUF.classes.aUFunitVanilla = AceOO.Class(aUF.classes.aUFunit)
aUF.classes.aUFunitCombo = AceOO.Class(aUF.classes.aUFunit)
aUF.classes.aUFunitFocus = AceOO.Class(aUF.classes.aUFunit)
aUF.classes.aUFunitMetro = AceOO.Class(aUF.classes.aUFunit)

function aUF.classes.aUFunit:ToString()
	return "aUFunit"
end

function aUF.classes.aUFunit.prototype:init(frame,unit)
	aUF.classes.aUFunit.super.prototype.init(self)
	
	self.name = frame:GetName()
	self.frame = frame
	self.frame.parent = self
	
	if self.frame and unit then
		self.frame:SetAttribute("unit", unit)
	else
		if self.frame.headerType then
			self.grouped = true
			self:TypeReset(self.frame.headerType)
		end
	end
end

function aUF.classes.aUFunit.prototype:AttributeChanged(name,value)
	if name ~= "unit" then return end
	
	if (value) then
		aUF:RegisterUnit(self, value)
	else
		aUF:UnregisterUnit(self)
	end

	if self.frame then
		if self.unit and self.unit ~= value then
			self.unit = value
			if self.unit then
				self:UpdateAll()
			else
				self:UpdateAll(true)
			end
		else
			self:UnitReset(value)
		end
	end
end

function aUF.classes.aUFunit.prototype:Reset(unitType)
    self:TypeReset(self.type)
    self:UnitReset(self.unit)
    self:UpdateAll()
end

function aUF.classes.aUFunit.prototype:TypeReset(unitType)
	if not unitType then
		return
	end
	self.type = unitType
	
	self.flags = {}
	self.disabledFrames = {}
	self.database = aUF.db.profile[self.type]
	
	self:SetAuraCount()
	self:SetupFrame()
	self:ApplyTheme()
end

function aUF.classes.aUFunit.prototype:UnitReset(unit)
	if not unit then return end
	
	self.unit = unit
	
	_, _, self.number = string.find(self.unit, "(%d+)")
	if not self.number then
		self.number = 1
	end
    
    local unitType = string.gsub(self.unit, "%d", "")
	if not self.type or self.type ~= unitType then
		self:TypeReset(unitType)
	end

	if string.find(self.unit,"pet") then
		if self.type == "partypet" or self.type == "raidpet" then
			self.parent = string.gsub(self.unit,"pet","")
		else
			self.parent = "player"
		end
	end
	
    self.eventsRegistered = false
	self:LoadScale()	
	self:BorderBackground()
	self:BarTexture()
    if not self.database.HideFrame then
        self.frame:Show()
    end
end


function aUF.classes.aUFunit.prototype:ObjectRegisterEvent(event,method)
    if type(method) == "function" then
        print(event)
        return
    end
    aUF:ObjectRegisterEvent(self, event, method)
end

function aUF.classes.aUFunit.prototype:ObjectUnregisterEvent(event,method)
    aUF:ObjectUnregisterEvent(self, event, method)
end

function aUF.classes.aUFunit.prototype:RegisterEvents(register)
	if register and self.eventsRegistered == false then	
		self:ObjectRegisterEvent("UNIT_HEALTH", "UpdateHealth")
		self:ObjectRegisterEvent("UNIT_MANA",	"UpdatePower")
		self:ObjectRegisterEvent("UNIT_RAGE", "UpdatePower")
		self:ObjectRegisterEvent("UNIT_FOCUS", "UpdatePower")
		self:ObjectRegisterEvent("UNIT_ENERGY", "UpdatePower")
		self:ObjectRegisterEvent("UNIT_DISPLAYPOWER","StatusBarsColor")
		
		self:ObjectRegisterEvent("UNIT_AURA", "UpdateAuras")
		self:ObjectRegisterEvent("UNIT_COMBAT","UnitCombat")
		self:ObjectRegisterEvent("UNIT_SPELLMISS","UnitSpellmiss")
        
		self:ObjectRegisterEvent("UPDATE_FACTION", "UpdatePvP")
		self:ObjectRegisterEvent("PLAYER_FLAGS_CHANGED", "UpdatePvP")
		self:ObjectRegisterEvent("UNIT_FACTION", "UpdatePvP")
		
		self:ObjectRegisterEvent("PARTY_LEADER_CHANGED", "LabelsCheckLeader")
		self:ObjectRegisterEvent("PARTY_LOOT_METHOD_CHANGED", "LabelsCheckLoot")		
		self:ObjectRegisterEvent("RAID_TARGET_UPDATE", "UpdateRaidTargetIcon")
		
		self:ObjectRegisterEvent("PLAYER_TARGET_CHANGED","UpdateHighlight")
        self:ObjectRegisterEvent("UNIT_PORTRAIT_UPDATE", "UpdatePortrait")
		self:ObjectRegisterEvent("PARTY_MEMBERS_CHANGED", "RosterUpdate")
		self:ObjectRegisterEvent("RAID_ROSTER_UPDATE", "RosterUpdate")
        
        if not self.inCombatSchedule then
            self.inCombatSchedule = self:ScheduleRepeatingEvent(self.UpdateStatusIcon, 0.5, self)
        end

        if self.ClassRegisterEvents then
            self:ClassRegisterEvents(true)
        end
        self.eventsRegistered = true
	elseif self.eventsRegistered == true and not register then
		local events = {
			"UNIT_HEALTH", "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_DISPLAYPOWER",
			"UNIT_AURA", "UNIT_COMBAT", "UNIT_SPELLMISS",
			"UPDATE_FACTION","PLAYER_FLAGS_CHANGED", "UNIT_FACTION",
			"PARTY_LEADER_CHANGED", "PARTY_LOOT_METHOD_CHANGED","RAID_TARGET_UPDATE",
			"PLAYER_TARGET_CHANGED","UNIT_PORTRAIT_UPDATE","RAID_ROSTER_UPDATE","PARTY_MEMBERS_CHANGED",
		}
		for k,v in pairs(events) do
			self:ObjectUnregisterEvent(v)
		end
        if self.inCombatSchedule then
			self:CancelScheduledEvent(self.inCombatSchedule)
			self.inCombatSchedule = nil
		end    
        if self.ClassRegisterEvents then
            self:ClassRegisterEvents(false)
        end
        self.eventsRegistered = false
	end    
end

-----------------
-- FRAME SETUP --
-----------------

function aUF.classes.aUFunit.prototype:SetupFrame()
	if self.frameSetup then return end
	self.ClassText = getglobal(self.name.."_Overlays_ClassText")
	self.NameLabel = getglobal(self.name.."_Overlays_NameLabel")
	self.HitIndicator = getglobal(self.name.."_Overlays_HitIndicator")
	self.HealthBar_BG = getglobal(self.name.."_HealthBar_BG")
	self.ManaBar_BG = getglobal(self.name.."_ManaBar_BG")
	self.XPBar_BG = getglobal(self.name.."_XPBar_BG")
	self.HealthBar = getglobal(self.name.."_HealthBar")
	self.HealthBar:SetScript("OnValueChanged",function() self:StatusBarsOnValueChanged(arg1) end)
	self.HealthBar:SetMinMaxValues(0,100)
	self.ManaBar = getglobal(self.name.."_ManaBar")
	self.ManaBar:SetMinMaxValues(0,100)
	self.XPBar = getglobal(self.name.."_XPBar")
	self.XPBar:SetMinMaxValues(0, 100)
	self.HealthText = getglobal(self.name.."_Overlays_HealthText")
	self.ManaText = getglobal(self.name.."_Overlays_ManaText")
	self.StatusText = getglobal(self.name.."_Overlays_StatusText")
	self.ComboText = getglobal(self.name.."_Overlays_ComboText")
	self.Highlight = getglobal(self.name.."_Highlights_Highlight")
	self.Highlight:SetAlpha(0.5)
	self.DebuffHighlight = getglobal(self.name.."_Highlights_Debuff")
	self.DebuffHighlight:SetAlpha(1)
	self.RaidTargetIcon = getglobal(self.name.."_Overlays_RaidTargetIcon")
	self.InCombatIcon = getglobal(self.name.."_Overlays_InCombatIcon")
	self.PVPIcon = getglobal(self.name.."_Overlays_PVPIcon")
	self.LeaderIcon = getglobal(self.name.."_Overlays_LeaderIcon")
	self.MasterIcon = getglobal(self.name.."_Overlays_MasterIcon")
	self.Highlights = getglobal(self.name.."_Highlights")
	self.Overlays = getglobal(self.name.."_Overlays")	
	self.Highlights:SetFrameLevel(self.Overlays:GetFrameLevel()+5)	
    
	self.frame:SetScript("OnEnter",function() self:OnEnter() end)
	self.frame:SetScript("OnLeave",function() self:OnLeave() end)
	self.frame:SetScript("OnDragStart",function() self:OnDragStart(arg1) end)
	self.frame:SetScript("OnDragStop",function() self:OnDragStop(arg1) end)
	self.frame:SetMovable(true)
	self.frame:EnableMouse(true)
	self.frame:RegisterForDrag("LeftButton")
	self.frame:RegisterForClicks("LeftButtonUp","RightButtonUp","MiddleButtonUp","Button4Up","Button5Up")
	self.frame:SetAttribute("type1", "target")
	self.frame:SetAttribute("*type2", "custom_menu")
	self.frame:SetAttribute("type2", "menu")
	self.frame.custom_menu = self.OnShowCustomMenu	
	self.frame.menu = self.OnShowMenu	
	
	ClickCastFrames = ClickCastFrames or {}
	ClickCastFrames[self.frame] = true
	
	self.frameSetup = true
end

function aUF.classes.aUFunit.prototype:CreatePortraitFrame()
	if self.Portrait then return self.Portrait end
	self.Portrait = CreateFrame("Button", self.name.."_Portrait", self.frame)
	self.Portrait:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 1, edgeFile = "", edgeSize = 0, insets = {left = 0, right = 0, top = 0, bottom = 0},})
	self.Portrait:EnableMouse(false)
	
	self.PortraitModel = getglobal(self.name.."_PortraitModel")
	self.PortraitModel:SetPoint("CENTER", self.Portrait, "CENTER", 0, 0)
	self.PortraitModel:SetScript("OnShow",function() this:SetCamera(0) end)
	self.Portrait:SetAlpha(self.frame:GetAlpha())
end

function aUF.classes.aUFunit.prototype:LoadScale()
	self.frame:SetScale(self.database.Scale)
end

function aUF.classes.aUFunit.prototype:LoadPosition()
	if self.grouped == true then return end
	self.frame:ClearAllPoints()
	if (aUF.db.profile.Positions[self.name]) then
        local x, y = strsplit(" ", aUF.db.profile.Positions[self.name])
		local scale = self.frame:GetEffectiveScale()
		self.frame:SetPoint("TOPLEFT", UIParent,"TOPLEFT", x/scale, y/scale)
	else
		self.frame:SetPoint("CENTER", UIParent, "CENTER")
	end
end

function aUF.classes.aUFunit.prototype:SavePosition()
	local scale = self.frame:GetEffectiveScale()
	local worldscale = UIParent:GetEffectiveScale()
	
	local x,y = self.frame:GetLeft()*scale,self.frame:GetTop()*scale - (UIParent:GetTop())*worldscale

    aUF.db.profile.Positions[self.name] = strjoin(" ", x, y)
end

function aUF.classes.aUFunit.prototype:ApplyTheme(resetWidth)
		local themetable = aUF.Layouts[self.database.FrameStyle]
		if not themetable then themetable = aUF.Layouts.ABF end
		
		self.disabledFrames = {}
		
		local height = 0
		local width = 0
		
		if self.database.HideMana == true then
			self.disabledFrames.ManaBar_BG = true
		else
			self.disabledFrames.ManaBar_BG = nil
		end

		if (self.unit ~= "player" and self.unit ~= "pet") or ((self.unit == "pet" or self.unit == "player") and self.database.ShowXP  == false) then
			self.disabledFrames.XPBar_BG = true
		else
			self.disabledFrames.XPBar_BG = nil
		end
		
		self.flags.ResizableBar = themetable.ResizableBar
		self.flags.BackgroundBarColor = themetable.BackgroundBarColor
		self.flags.AlphaBar = themetable.AlphaBar
		self.flags.RaidColorName = themetable.RaidColorName
		self.flags.ThemeName = themetable.Name
		self.flags.PetClassName = themetable.PetClassName
		self.flags.HappinessBar = themetable.HappinessBar
		self.flags.Portrait = themetable.Portrait
		
		if self.database.Portrait and self.flags.Portrait then
			self.disabledFrames.Portrait = nil
			if not self.Portrait then
				self:CreatePortraitFrame()
			end
		else
			self.disabledFrames.Portrait = true
			self.flags.Portrait = false
		end
		
		if self.unit == "target" then
			if themetable.ComboGFX == true then
				comboGFX = 1
			else
				comboGFX = 0
			end
		end
		
		height = themetable.ThemeData.all.FrameHeight
		width = themetable.ThemeData.all.FrameWidth
		
		if themetable.ThemeData[self.type] then
			if themetable.ThemeData[self.type].FrameHeight then
				height = themetable.ThemeData[self.type].FrameHeight
			end
			if themetable.ThemeData[self.type].FrameWidth then
				width = themetable.ThemeData[self.type].FrameWidth	
			end
		end
		
		if manabarhide == true then
			self.flags.RaidColorName = true
		end
		
		for k,v in pairs(themetable.ThemeData.all) do
			if self[k] then self[k]:ClearAllPoints() end
		end
		local index = "all"
		for j = 1,2 do
			if j == 2 then
				index = self.type
			end
			if themetable.ThemeData[index] then
				for key, value in pairs(themetable.ThemeData[index]) do
					if self[key] and value.Hidden ~= true and not (self.disabledFrames[key]) then
						local useSecondary = false
						
						if value.Hide then
							self.disabledFrames[value.Hide] = true
						end
						
						if value.HeightAdd then
							height = height + value.HeightAdd
						end

						if value.Font then
							self[key]:SetFont(value.Font,value.FontSize)
						end
						
						local RelativeTo
						if value.RelativeTo then
							if themetable.ThemeData[index][value.RelativeTo]
							and (themetable.ThemeData[index][value.RelativeTo].Hidden == true 
							or not getglobal(self.name.."_"..value.RelativeTo) 
							or self.disabledFrames[value.RelativeTo]) 
							and value.RelativeTo2 then
								if themetable.ThemeData[index][value.RelativeTo2]
								and (themetable.ThemeData[index][value.RelativeTo2].Hidden == true 
								or not getglobal(self.name.."_"..value.RelativeTo2) 
								or self.disabledFrames[value.RelativeTo2]) 
								and value.RelativeTo3 then
									RelativeTo = self[value.RelativeTo3]
									useSecondary = 3
								else
									RelativeTo = self[value.RelativeTo2]
									useSecondary = 2
								end
							else
								RelativeTo = self[value.RelativeTo]								
							end
						elseif not value.RelativeTo or not getglobal(self.name.."_"..value.RelativeTo)  then
							RelativeTo = self.name
						end
						
						if value.Width then
							if self.flags.Portrait == true and value.WidthP then
								self[key]:SetWidth(value.WidthP)
							elseif useSecondary and value["Width"..useSecondary] then
								self[key]:SetWidth(value["Width"..useSecondary])
							else
								self[key]:SetWidth(value.Width)
							end
						end
						
						if value.Height then
							self[key]:SetHeight(value.Height)
						end
						
						local x,y,Point,RelativePoint
						local positionSet = true
						if value.x and value.y and value.Point and value.RelativePoint and not useSecondary then
							x = value.x
							y = value.y
							Point = value.Point
							RelativePoint = value.RelativePoint
						elseif value.x2 and value.y2 and value.Point2 and value.RelativePoint2 and useSecondary == 2 then
							x = value.x2
							y = value.y2
							Point = value.Point2
							RelativePoint = value.RelativePoint2
						elseif value.x3 and value.y3 and value.Point3 and value.RelativePoint3 and useSecondary == 3 then
							x = value.x3
							y = value.y3
							Point = value.Point3
							RelativePoint = value.RelativePoint3			
						else
							positionSet = false
						end
						
						if positionSet == true then
							local pointAdjust
							if themetable.ResizableBar == true and self.database.LongStatusbars == false and (key == "HealthText" or key == "ManaText") then
								if Point == "RIGHT" then 
									pointAdjust = "LEFT"
								elseif Point == "LEFT" then 
									pointAdjust = "RIGHT"
								else
									pointAdjust = value.Point
								end
							else
								pointAdjust = Point
							end
							self[key]:ClearAllPoints()
							self[key]:SetPoint(pointAdjust, RelativeTo, RelativePoint, x, y)
							if value.Endx and value.Endy and value.EndPoint and value.EndRelativePoint and value.EndRelativeTo then
								self[key]:SetPoint(value.EndPoint, self[value.EndRelativeTo], value.EndRelativePoint, value.Endx, value.Endy)
							end
						end
						
						if value.Visibility then
							for k, v in pairs(value.Visibility) do
								if self[v] and ((themetable.ThemeData[index][v] and not (themetable.ThemeData[index][v].Hidden or self.disabledFrames[v]))or (not themetable.ThemeData[index][v])) then
									self[v]:Show()
								end
							end
						end
						if value.Justify then
							self[key]:SetJustifyH(value.Justify)
						end
					elseif self[key] and (value.Hidden or self.disabledFrames[key] == true ) then
						self[key]:Hide()
						if value.Visibility then
							for k, v in pairs(value.Visibility) do
								if self[v] then
									self.disabledFrames[v] = true
									self[v]:Hide()
								end
							end
						end
					end
				end
			end
		end
		for key,value in pairs(self.disabledFrames) do
			if self[key] then
				self[key]:Hide()
			end
		end
	
	self.frame:SetHeight(height)
	self.frame.height = height
	aUF.heightTable[self.type] = height
   
    self.frame:SetWidth(100)
      
    if resetWidth and self.widthAdd then
        self.database.Width = width
    end
    
    if self.flags.ResizableBar == true and self.database.LongStatusbars == false then
        self.HealthBar_BG:SetWidth(self.HealthBar_BG:GetWidth() - self.HealthText:GetWidth())
        self.ManaBar_BG:SetWidth(self.ManaBar_BG:GetWidth() - self.ManaText:GetWidth())
    end
    
    self.widthAdd = 0
    self:SetWidth()
   
	if self.flags.Portrait then
		self.PortraitModel:SetWidth(self.Portrait:GetWidth()-2)
		self.PortraitModel:SetHeight(self.Portrait:GetHeight()-2)
	end
end

function aUF.classes.aUFunit.prototype:BorderBackground()
	local colortable
	local bordercolor = aUF.db.profile.FrameBorderColors
	local borderstyle = aUF.db.profile.BorderStyle
	
	if self.unit == "target" then
		colortable = aUF.db.profile.TargetFrameColors
	else
		colortable = aUF.db.profile.PartyFrameColors
	end
	
	self.frame:SetBackdrop({
				bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
				edgeFile = aUF.Borders[borderstyle].texture, edgeSize = aUF.Borders[borderstyle].size,
				insets = {left = aUF.Borders[borderstyle].insets, right = aUF.Borders[borderstyle].insets, top = aUF.Borders[borderstyle].insets, bottom = aUF.Borders[borderstyle].insets},
		})
	
	self.frame:SetBackdropColor(colortable.r,colortable.g,colortable.b,colortable.a)
	self.frame:SetBackdropBorderColor(bordercolor.r,bordercolor.g,bordercolor.b,bordercolor.a)

	if self.flags.Portrait then
		self.Portrait:SetBackdropBorderColor(self.frame:GetBackdropBorderColor())
		self.Portrait:SetBackdropColor(self.frame:GetBackdropColor())
	end
end

function aUF.classes.aUFunit.prototype:SetWidth()
	local widthAdd = 100
	if self.database.Width then widthAdd = self.database.Width end
	if not self.widthAdd then self.widthAdd = 0 end
	widthAdd = widthAdd - 100
	
	widthAdd =  widthAdd - self.widthAdd
	self.widthAdd = self.widthAdd + widthAdd
	
	self.frame:SetWidth(self.frame:GetWidth() + widthAdd)
	self.HealthBar_BG:SetWidth(self.HealthBar_BG:GetWidth() + widthAdd)
	self.ManaBar_BG:SetWidth(self.ManaBar_BG:GetWidth() + widthAdd)
	self.XPBar_BG:SetWidth(self.XPBar_BG:GetWidth() + widthAdd)
	
	if self.XPBar_Rest then
		self.XPBar_Rest:ClearAllPoints()
		self.XPBar_Rest:SetPoint("LEFT",self.XPBar_BG,"LEFT")
		self.XPBar_Rest:SetHeight(self.XPBar_BG:GetHeight())
		self.XPBar_Rest:SetWidth(self.XPBar_BG:GetWidth())
	end
	if self.XPBar then
		self.XPBar:ClearAllPoints()
		self.XPBar:SetPoint("LEFT",self.XPBar_BG,"LEFT")
		self.XPBar:SetHeight(self.XPBar_BG:GetHeight())
		self.XPBar:SetWidth(self.XPBar_BG:GetWidth())
	end
	if self.HealthBar then
		self.HealthBar:ClearAllPoints()
		self.HealthBar:SetPoint("LEFT",self.HealthBar_BG,"LEFT")
		self.HealthBar:SetHeight(self.HealthBar_BG:GetHeight())
		self.HealthBar:SetWidth(self.HealthBar_BG:GetWidth())
	end
	if self.ManaBar then
		self.ManaBar:ClearAllPoints()
		self.ManaBar:SetPoint("LEFT",self.ManaBar_BG,"LEFT")
		self.ManaBar:SetHeight(self.ManaBar_BG:GetHeight())
		self.ManaBar:SetWidth(self.ManaBar_BG:GetWidth())
	end
	
	self:AuraPosition()
end

function aUF.classes.aUFunit.prototype:BarTexture()
	local barstyle = aUF.db.profile.BarStyle
	
	self.HealthBar:SetStatusBarTexture(SurfaceLib:Fetch(barstyle))
	self.ManaBar:SetStatusBarTexture(SurfaceLib:Fetch(barstyle))
	self.XPBar:SetStatusBarTexture(SurfaceLib:Fetch(barstyle))
	if self.XPBar_Rest then
		self.XPBar_Rest:SetStatusBarTexture(SurfaceLib:Fetch(barstyle))
	end
	self.HealthBar_BG:SetTexture(SurfaceLib:Fetch(barstyle))
	self.ManaBar_BG:SetTexture(SurfaceLib:Fetch(barstyle))
	self.XPBar_BG:SetTexture(SurfaceLib:Fetch(barstyle))
end

function aUF.classes.aUFunit.prototype:SetUnitWatch()
	if self.database.HideFrame == true then
		self.watched = false
		UnregisterUnitWatch(self.frame)
		self.frame:Hide()
	else
		RegisterUnitWatch(self.frame, false)
		self.watched = true
	end
end

-------------
-- UPDATES --
-------------

function aUF.classes.aUFunit.prototype:UpdateHighlight(entered)
	if (UnitExists("target") and UnitIsUnit("target",self.unit) and not (string.find(self.unit,"target")) and aUF.db.profile.HighlightSelected == true) or (entered == true and aUF.db.profile.HighlightSelected == true) then
		self.Highlight:Show()
	else
		self.Highlight:Hide()
	end
end

function aUF.classes.aUFunit.prototype:UpdatePetAll()
	if arg1 ~= self.parent then return end
--	self:UpdateAll()
end

function aUF.classes.aUFunit.prototype:UpdateAll(unload,highlight)
	if not self.database or not self.unit then return end
    if not InCombatLockdown() then
		self:SetUnitWatch()
	end
	if not unload and not self.database.HideFrame then
		if not InCombatLockdown() then
			self:LoadPosition()
		end
		self:StatusBarsColor()
		self:UpdateHealth()
		self:UpdatePower()
		self:UpdatePvP()
		self:UpdateAuras()
		self:UpdateRaidTargetIcon()
		self:UpdateStatusIcon()
		self:LabelsCheckLeader()
        if not highlight then
            self:UpdateHighlight()
        end
        self:GetStringEvents()
		self:UpdateTextStrings()
        self:UpdatePortrait()
		self:RegisterEvents(true)
	else
        self:GetStringEvents(true)            
		self:RegisterEvents()
	end
end

function aUF.classes.aUFunit.prototype:UpdatePortrait(stop)
    if not self.flags.Portrait then
        if self.PortraitModel then
            self.PortraitModel:Hide()
        end
        return
    end
	if not UnitExists(self.unit) or not UnitIsConnected(self.unit) or not UnitIsVisible(self.unit) then
		self.PortraitModel:SetModelScale(4.25)
		self.PortraitModel:SetPosition(0,0,-1.5)
		self.PortraitModel:SetModel("Interface\\Buttons\\talktomequestionmark.mdx")
	else
		self.PortraitModel:SetUnit(self.unit)
		self.PortraitModel:SetCamera(0)
		self.PortraitModel:Show()
	end
	self.Portrait:SetBackdropBorderColor(self.frame:GetBackdropBorderColor())
	self.Portrait:SetBackdropColor(self.frame:GetBackdropColor())
end

function aUF.classes.aUFunit.prototype:UpdateTextStrings()
	local strings = {"HealthText","ManaText","NameLabel","ClassText"}
    for k,v in pairs(strings) do
        local str = self[v]
        if str and str.func then
            str.func(self.unit,str)
        end
    end
end

function aUF.classes.aUFunit.prototype:GetStringEvents(unreg)
	local strings = {"HealthText","ManaText","NameLabel","ClassText"}
	for k,v in pairs(strings) do
        if aUF.HelperFunctions[self.type] and aUF.HelperFunctions[self.type][v] then
            local str = self[v]
            if unreg and str then
                if str.events then
                    for k,v in pairs(str.events) do
                        self:ObjectUnregisterEvent(v,k)
                    end
                end
                str.events = {}
            end
            if not unreg and str then
                local events = aUF.HelperFunctions[self.type][v].events
                if events then
                    for x,y in pairs(events) do
                        self:StringRegisterEvent(str, y, v)
                        str.func = aUF.HelperFunctions[self.type][v].func
                    end
                end
            end
        end
	end
end

function aUF.classes.aUFunit.prototype:StringRegisterEvent(str, event, shortName)
    if not str.events then
        str.events = {}
    end
    local unit = self.unit
    self[shortName..event] = function() str.func(unit,str) end
    str.events[shortName..event] = event
    self:ObjectRegisterEvent(event,shortName..event)
end

---------------
-- STATUSBAR --
---------------

function aUF.classes.aUFunit.prototype:UpdateHealth()
	local currValue,maxValue = UnitHealth(self.unit),UnitHealthMax(self.unit)
    local perc = currValue/maxValue * 100
	
	if ( not UnitExists(self.unit) or UnitIsDead(self.unit) or UnitIsGhost(self.unit) or not UnitIsConnected(self.unit) or maxValue == 1) then
		self.HealthBar:SetValue(0)
	else
		self.HealthBar:SetValue(perc)
	end
end

function aUF.classes.aUFunit.prototype:UpdatePower()
	local currValue,maxValue = UnitMana(self.unit),UnitManaMax(self.unit)
    local perc = currValue/maxValue * 100	
	local db = aUF.db.profile
	
	if ( not UnitExists(self.unit) or UnitIsDead(self.unit) or UnitIsGhost(self.unit) or not UnitIsConnected(self.unit) or (currValue == 1 and maxValue == 1) ) then
		self.ManaBar:SetValue(0)
	else
		if maxValue == 0 then
			self.ManaBar:SetValue(0)
		else
			self.ManaBar:SetValue(perc)
		end
	end
	
	if not self.powertype or self.powertype ~= UnitPowerType(self.unit) then
		self:StatusBarsColor()
	end
end

function aUF.classes.aUFunit.prototype:StatusBarsColor()
	local db = aUF.db.profile
	local healthColor
	local _,Eclass = UnitClass(self.unit)
	
	if string.find(self.unit,"target") and not UnitIsFriend(self.unit, "player") and (aUF.db.profile.TargetShowHostile == true ) then
		healthColor = aUF:UtilFactionColors(self.unit)
		if aUF.db.profile.SmoothHealthBars == true then
			self.flags.BarColor = 2
		else
			self.flags.BarColor = nil
		end		
	elseif self.unit == "pet" and self.flags.HappinessBar == true and Eclass ~= "WARLOCK" then
		local happiness = GetPetHappiness()
		if ( happiness == 1 ) then
			healthColor = aUF.ManaColor[1]
		elseif ( happiness == 2 ) then
			healthColor = aUF.ManaColor[3]
		else
			healthColor = db.HealthColor
		end
		self.flags.BarColor = nil
	elseif UnitIsPlayer(self.unit) and self.database.ClassColorBars == true and RAID_CLASS_COLORS[Eclass] then
		healthColor = RAID_CLASS_COLORS[Eclass]
		self.flags.BarColor = nil
	else
		healthColor = db.HealthColor
		if aUF.db.profile.SmoothHealthBars == true then
			self.flags.BarColor = 1
		else
			self.flags.BarColor = nil
		end
	end
	
	if ( self.flags.BarColor ) then
		self:StatusBarsOnValueChanged(self.HealthBar:GetValue())
	else
		self.HealthBar:SetStatusBarColor(healthColor.r,healthColor.g,healthColor.b, 1)
	end
	
	if self.flags.BackgroundBarColor == true then
		self.HealthBar_BG:SetVertexColor(healthColor.r,healthColor.g,healthColor.b,0.25)
	else
		self.HealthBar_BG:SetVertexColor(0,0,0,0.25)
	end

	-- mana
	self.powertype = UnitPowerType(self.unit)
	local info = db.ManaColor[self.powertype]
	if self.flags.AlphaBar == true then
		self.ManaBar_BG:SetVertexColor(info.r,info.g,info.b,0.25)
	else
		self.ManaBar_BG:SetVertexColor(0,0,0,0.25)
	end
	self.ManaBar:SetStatusBarColor(info.r,info.g,info.b,0.8)
end

function aUF.classes.aUFunit.prototype:StatusBarsOnValueChanged(value)
	if self.flags.BarColor == 1 then
		self:SmoothHealthBarOnValueChanged(value)
	elseif self.flags.BarColor == 2 then
		self:SmoothTargetHealthBarOnValueChanged(value,aUF:UtilFactionColors(self.unit))
	end
end

function aUF.classes.aUFunit.prototype:SmoothTargetHealthBarOnValueChanged(value,colortable)
	if ( not value) then
		return
	end
	local r, g, b
	local min, max = self.HealthBar:GetMinMaxValues()
	if ( (value < min) or (value > max) ) then
		return
	end
	if ( (max - min) > 0 ) then
		value = (value - min) / (max - min)
	else
		value = 0
	end
	
	r = colortable.r*(0.35*value+0.65)
	g = colortable.g*(0.35*value+0.65)
	b = colortable.b*(0.35*value+0.65)
	self.HealthBar:SetStatusBarColor(r, g, b)
end
	
function aUF.classes.aUFunit.prototype:SmoothHealthBarOnValueChanged(value)
	if ( not value ) then
		return
	end
	local db = aUF.db.profile
	local r, g, b
	local min, max = self.HealthBar:GetMinMaxValues()
	if ( (value < min) or (value > max) ) then
		return
	end
	if ( (max - min) > 0 ) then
		value = (value - min) / (max - min)
	else
		value = 0
	end
	if(value > 0.5) then
		r = (db.HealthColor.r) + (((1-value) * 2)* (1-(db.HealthColor.r)))
		g = db.HealthColor.g
	else
		r = 1.0
		g = (db.HealthColor.g) - (0.5 - value)* 2 * (db.HealthColor.g)
	end
	b = db.HealthColor.b
	self.HealthBar:SetStatusBarColor(r, g, b)
end

-----------
-- ICONS --
-----------

function aUF.classes.aUFunit.prototype:RosterUpdate()
    self:LabelsCheckLeader()
    self:LabelsCheckLoot()
    self:UpdateRaidTargetIcon()
end

function aUF.classes.aUFunit.prototype:LabelsCheckLeader()
	self.LeaderIcon:Hide()
	if not ( self.type == "raid" or self.type == "party" or self.unit == "player") then return end
	if aUF.db.profile.ShowGroupIcons == true then
		if self.type == "raid" then
			local _, rank = GetRaidRosterInfo(self.number)
			if rank == 2 then
				self.LeaderIcon:Show()
			end
		elseif self.type == "party" and tonumber(GetPartyLeaderIndex()) == tonumber(self.number) then
			self.LeaderIcon:Show()
		elseif self.unit == "player" and IsPartyLeader() then
			self.LeaderIcon:Show()
		end
	end
end

function aUF.classes.aUFunit.prototype:LabelsCheckLoot()
	self.MasterIcon:Hide()
	if not ( self.unit == "player" or self.type == "party") then return end
	if aUF.db.profile.ShowGroupIcons == true then
		local _, lootMaster = GetLootMethod()
		if lootMaster then
			if self.unit == "player" and lootMaster == 0 then
				self.MasterIcon:Show()
			elseif self.unit ~= "player" and lootMaster > 0 then
				if lootMaster == self.number and not (string.find(self.unit,"pet")) then
					self.MasterIcon:Show()
				end
			end
		end
	end
end

function aUF.classes.aUFunit.prototype:UpdateRaidTargetIcon()
	local index = GetRaidTargetIndex(self.unit)
	if ( index ) and aUF.db.profile[self.type].ShowRaidTargetIcon == true then
		SetRaidTargetIconTexture(self.RaidTargetIcon, index)
		self.RaidTargetIcon:Show()
	else
		self.RaidTargetIcon:Hide()
	end
end

function aUF.classes.aUFunit.prototype:UpdateStatusIcon()
    if self.database.ShowInCombatIcon == true then
        if UnitAffectingCombat(self.unit) then
            self.InCombatIcon:Show()
            self.InCombatIcon:SetTexCoord(0.5,1,0,0.5)
        elseif self.unit == "player" and IsResting() then
            self.InCombatIcon:Show()
            self.InCombatIcon:SetTexCoord(0,0.5,0,0.421875)
        else
            self.InCombatIcon:Hide()
        end
    else
        self.InCombatIcon:Hide()
    end
end

function aUF.classes.aUFunit.prototype:UpdatePvP()
	if aUF.db.profile.ShowPvPIcon == true then
		if ( UnitIsPVPFreeForAll(self.unit) ) then
			self.PVPIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-FFA")
			self.PVPIcon:Show()
		elseif ( UnitFactionGroup(self.unit) and UnitIsPVP(self.unit) ) then
			self.PVPIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..UnitFactionGroup(self.unit))
			self.PVPIcon:Show()
		else
			self.PVPIcon:Hide()
		end
	else
		self.PVPIcon:Hide()
	end
end

-----------
-- AURAS --
-----------

function aUF.classes.aUFunit.prototype:SetAuraCount()	
	local db = self.database
	self.AuraRows, self.AuraColumns = db.AuraRows, db.AuraColumns

	local count = self.AuraRows * self.AuraColumns
	
	-- setup auras
	for i = 1,math.max(count, 20) do
		if (not self["Aura"..i]) then
			self["Aura"..i] = getglobal(self.name.."_Aura"..i) or CreateFrame("Button", self.name .. "_Aura" .. i, self.frame, "AGAuraTemplate")
			
			self["Aura"..i].Icon = getglobal(self.name.."_Aura"..i.."Icon")
			self["Aura"..i].Overlay = getglobal(self.name.."_Aura"..i.."Overlay")
			self["Aura"..i].Count = getglobal(self.name.."_Aura"..i.."Count")
			self["Aura"..i]:SetScript("OnEnter",function() self:AuraOnEnter() end)
			self["Aura"..i]:SetScript("OnLeave",function() GameTooltip:Hide() end)

			aUF.hooks.hookShowHide(self["Aura"..i])
			aUF.hooks.hookSetTexture(self["Aura"..i].Icon)
                                
--			self["Aura"..i]:SetParent(self.Overlays)
			self["Aura"..i]:Hide()
		end
	end
	
	local i = count+1
	while (true) do
		local f = self["Aura"..i]
		if (not f) then
			break
		end
		f:Hide()
		i = i+1
	end
end

function aUF.classes.aUFunit.prototype:AuraPosition()
	local position = self.database.AuraPos	
	local count = self.AuraRows * self.AuraColumns
	
	self.Aura1:ClearAllPoints()
	
    local Point, RelativePoint,x,y,GrowDirection,Scale,Alpha,AdjustAuto
--  local auraTable = aUF.Layouts[self.database.FrameStyle].Auras.all
    local auraTable
    if aUF.Layouts[self.database.FrameStyle] then
        auraTable = aUF.Layouts[self.database.FrameStyle].Auras or aUF.Layouts.ABF.Auras
    else
        return
    end

    if auraTable[position] then
        Point = auraTable[position].Point
        RelativePoint = auraTable[position].RelativePoint
        x = auraTable[position].x
        y = auraTable[position].y
        GrowDirection = auraTable[position].GrowDirection
        Scale = auraTable[position].Scale
        Alpha = auraTable[position].Alpha
        AdjustAuto = auraTable[position].AdjustAuto
    else
        return
    end   
    
    if Scale == "auto" then
        Scale = (self.frame:GetWidth()-AdjustAuto)/(16*self.AuraColumns)
        self.Aura1:SetPoint(Point,self.frame,RelativePoint, x*(1/Scale), y*(1/Scale))
    else
        self.Aura1:SetPoint(Point,self.frame,RelativePoint, x*(1/Scale), y*(1/Scale))
    end
    
    
    for i=1,count do
		self["Aura"..i]:SetScale(Scale)
		self["Aura"..i]:SetAlpha(Alpha)
	end

	if ( GrowDirection == "left" ) then
		for i=2,count do
			self["Aura"..i]:ClearAllPoints()
			if (math.fmod(i,self.AuraColumns) == 1) then
				self["Aura"..i]:SetPoint("TOP",self["Aura"..i-self.AuraColumns],"BOTTOM",0,-1)
			else
				self["Aura"..i]:SetPoint("RIGHT",self["Aura"..i-1],"LEFT",1,0)
			end
		end
	else
		for i=2,count do
			self["Aura"..i]:ClearAllPoints()
			if (i % self.AuraColumns == 1) then
				if position == "Above" then
					self["Aura"..i]:SetPoint("BOTTOM",self["Aura"..i-self.AuraColumns],"TOP",0,1)
 				else
					self["Aura"..i]:SetPoint("TOP",self["Aura"..i-self.AuraColumns],"BOTTOM",0,-1)
				end
			else
				self["Aura"..i]:SetPoint("LEFT",self["Aura"..i-1],"RIGHT",1,0)
			end
		end
	end
end

function aUF.classes.aUFunit.prototype:UpdateAuras()	
	if aUF.auraUpdates > 4 then 
		aUF.auraUpdatePool[self.name] = true
--		print(self.name.." add to pool")
		return 
	else
--		print(self.name.." update "..aUF.auraUpdates)
		aUF.auraUpdatePool[self.name] = nil 
		aUF.auraUpdates = aUF.auraUpdates + 1
	end
	
	local filter,position,AuraDebuffC = self.database.AuraFilter,self.database.position,self.database.AuraDebuffC
    local position = self.database.AuraPos
	local dbcount,bcount,lastDebuff = 0,0,0
	local buttons,dFound
	local rows, cols, auracount = self.AuraRows, self.AuraColumns, self.AuraRows * self.AuraColumns;
	
	self.DebuffHighlight:Hide()
	local hide = position == "Hidden"
	local hidef = self["Aura1"].Hide
	
	-- hide all auras
	if (not self.hidden) then
		for i=1,auracount do
			hidef(self["Aura"..i])
		end
			
		self.hidden = true
	end

	if (not self.unit) then
		return
	else
		-- look for debuffs
		-- note: not using aUF:UnitBuff, as the third argument is showDispellable
		local i = 1
		local debuff, t, highlight
		repeat
			_, _, debuff, _, t = UnitDebuff(self.unit,i,filter)
			if (debuff) then dbcount = dbcount + 1 end
			if (not highlight and t and DebuffTypeColor[t]) then
				highlight = DebuffTypeColor[t]
			end
			i = i + 1
		until (not debuff)
		
		-- set highlight
		if (highlight and AuraDebuffC == true and UnitIsFriend("player",self.unit)) then
			self.DebuffHighlight:Show()
			self.DebuffHighlight:SetVertexColor(highlight.r, highlight.g, highlight.b)
		end
		
		i = 1
		repeat
			debuff = UnitBuff(self.unit, i, filter)
			if (debuff) then bcount = bcount + 1 end
			i = i + 1
		until (not debuff)
	end
	
	if hide or bcount + dbcount == 0 then
		self.hidden = true
		return
	end
	self.hidden = false
    
	if (auracount < bcount + dbcount) then
		if (self.database.AuraPreferBuff) then
			dbcount = auracount - bcount
			if (dbcount < 0) then
				bcount = auracount
				dbcount = 0;
			end
		else
			bcount = auracount - dbcount
			if (bcount < 0) then
				dbcount = auracount;
				bcount = 0;
			end
		end
	end
	
	local buff, count, class, index;
	
	for i=1,bcount do
		_, _, buff, count = UnitBuff(self.unit,i,filter)
		self:SetAura(i, false, i, buff, count, nil)
	end
	
	rows = math.ceil((bcount+dbcount) / cols);
	
	if (rows == 1 and (position == "Left" or position == "Right" or position == "Inside")) then
		if (cols > bcount+dbcount) then
			cols = bcount+dbcount+0
		end
	end
	
	for i=1,dbcount do
		_,_, buff, count, class = UnitDebuff(self.unit,i,filter)
		index = rows * cols - i + 1
		self:SetAura(index, true, i, buff, count, class)
	end
end

function aUF.classes.aUFunit.prototype:SetAura(index, isDebuff, id, buff, count, class)
	local buffFrame = self["Aura" .. index]
	buffFrame.Icon:SetTexture(buff)
	buffFrame:Show()
	buffFrame.id = id
    
	if count and count > 1 then
		buffFrame.Count:SetText(count)
	else
		buffFrame.Count:SetText("")
	end

	if (isDebuff) then
		buffFrame.buffFilter = "HARMFUL"
		local borderColor;
		if class then
			borderColor = DebuffTypeColor[class]
		else
			borderColor = DebuffTypeColor["none"]
		end
		buffFrame.Overlay:SetVertexColor(borderColor.r, borderColor.g, borderColor.b)
	buffFrame.Overlay:Show()
	else
		buffFrame.buffFilter = "HELPFUL"
		buffFrame.Overlay:Hide()
	end
end

function aUF.classes.aUFunit.prototype:ShowAuraPositions()
	local count = self.AuraRows*self.AuraColumns;
	if (self.database.AuraPreferBuff) then
		for i = 1,count do
			local b = self["Aura"..i];
			if (i <= 16) then
				b.Icon:SetTexture("Interface\\Icons\\Spell_ChargePositive");
			else
				b.Icon:SetTexture("Interface\\Icons\\Spell_ChargeNegative");
			end
		end
	else
		for i = 1,count do
			local b = self["Aura"..i];
			if (i > count-16) then
				b.Icon:SetTexture("Interface\\Icons\\Spell_ChargeNegative");
			else
				b.Icon:SetTexture("Interface\\Icons\\Spell_ChargePositive");
			end
		end
	end
	for i = 1,count do
		local b = self["Aura"..i];
		b:Show();
		b.Overlay:Show();
	end
	
	self.hidden = false;
	self:ScheduleEvent(function () self:UpdateAuras() end, 10);
end

function aUF.classes.aUFunit.prototype:AuraOnEnter()
	if (not this:IsVisible()) then return end
	GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
	if ( this.buffFilter == "HELPFUL") then
		GameTooltip:SetUnitBuff(self.unit, this.id, self.database.AuraFilter)
	elseif ( this.buffFilter == "HARMFUL") then
		GameTooltip:SetUnitDebuff(self.unit, this.id, 0)
	end
end

-----------------------
-- MOUSE INTERACTION --
-----------------------

function aUF.classes.aUFunit.prototype:OnDragStart(button)
	if self.grouped then
		self.frame:GetParent().object:OnDragStart(button)
		return
	end
	if button == "LeftButton" and (aUF.db.profile.Locked == false or IsAltKeyDown()) then
		self.frame:StartMoving()
	end
end

function aUF.classes.aUFunit.prototype:OnDragStop(button)
	if self.grouped then
		self.frame:GetParent().object:OnDragStop(button)
		return
	end
	self.frame:StopMovingOrSizing()
	self:SavePosition()
	self.frame:SetUserPlaced(false)
end

function aUF.classes.aUFunit.prototype:OnShowMenu(unit)
    self.parent:DropDownUnit(self.unit)
end

function aUF.classes.aUFunit.prototype:OnShowCustomMenu(unit)
	if not InCombatLockdown() then
		aUF.dewdrop:Open(self, 'children', aUF:CreateDewdrop(self.parent.type),'cursorX', true, 'cursorY', true)
	else
		aUF:Print(L["lockdown"])	
	end
end

function aUF.classes.aUFunit.prototype:OnEnter()
	self.frame.unit = self.unit
	self:UpdateHighlight(true)
	UnitFrame_OnEnter()
end

function aUF.classes.aUFunit.prototype:OnLeave()
	self:UpdateHighlight()
	UnitFrame_OnLeave()
end

function aUF.classes.aUFunit.prototype:DropDownUnit()
	local type = nil

	if self.unit == "player" then
		type = PlayerFrameDropDown
	elseif self.unit == "target" then
		type = TargetFrameDropDown
	elseif self.unit == "pet" then
		type = PetFrameDropDown
	elseif self.type == "party" then
		type = getglobal("PartyMemberFrame"..self.number.."DropDown")
	elseif string.find(self.unit, "raid") then
		type = FriendsDropDown
		this.unit = self.unit
		this.name = UnitName(self.unit)
		this.id = this:GetID()
		FriendsDropDown.displayMode = "MENU"
		FriendsDropDown.initialize = RaidFrameDropDown_Initialize
	end

	if self.number then this:SetID(self.number) end

	if type then
		HideDropDownMenu(1)
		type.unit = self.unit
		type.name = UnitName(self.unit)
		ToggleDropDownMenu(1, nil, type,"cursor")
		return true
	end
end

-------------------
-- COMBAT EVENTS --
-------------------

function aUF.classes.aUFunit.prototype:UnitCombat(unit)
	if unit ~= self.unit then return end
	if not ( self.database.ShowCombat ) then return end
	self:CombatFeedback_OnCombatEvent(arg1, arg2, arg3, arg4, arg5)
end

function aUF.classes.aUFunit.prototype:UnitSpellmiss(unit)
	if unit ~= self.unit then return end
	if not ( self.database.ShowCombat ) then return end
	self:CombatFeedback_OnSpellMissEvent(arg1, arg2)
end

function aUF.classes.aUFunit.prototype:CombatFeedback_OnCombatEvent(unit, event, flags, amount, type)
	local fontHeight = 13
	local text = ""
	local r,g,b = 1,1,1
	if( event == "IMMUNE" ) then
		fontHeight = fontHeight * 0.75
		text = CombatFeedbackText[event]
	elseif ( event == "WOUND" ) then
		if ( amount ~= 0 ) then
			if ( flags == "CRITICAL" or flags == "CRUSHING" ) then
				fontHeight = fontHeight * 1.5
			elseif ( flags == "GLANCING" ) then
				fontHeight = fontHeight * 0.75
			end
			if ( type > 0 ) then
				r = 1.0
				g = 1.0
				b = 0.0
			end
			if UnitInParty(self.unit) or UnitInRaid(self.unit) then
				r = 1.0
				g = 0.0
				b = 0.0
			end
			text = "-"..amount
		elseif ( flags == "ABSORB" ) then
			fontHeight = fontHeight * 0.75
			text = CombatFeedbackText["ABSORB"]
		elseif ( flags == "BLOCK" ) then
			fontHeight = fontHeight * 0.75
			text = CombatFeedbackText["BLOCK"]
		elseif ( flags == "RESIST" ) then
			fontHeight = fontHeight * 0.75
			text = CombatFeedbackText["RESIST"]
		else
			text = CombatFeedbackText["MISS"]
		end
	elseif ( event == "BLOCK" ) then
		fontHeight = fontHeight * 0.75
		text = CombatFeedbackText[event]
	elseif ( event == "HEAL" ) then
		text = "+"..amount
		r = 0.0
		g = 1.0
		b = 0.0
		if ( flags == "CRITICAL" ) then
			fontHeight = fontHeight * 1.3
		end
	elseif ( event == "ENERGIZE" ) then
		text = amount
		r = 0.41
		g = 0.8
		b = 0.94
		if ( flags == "CRITICAL" ) then
			fontHeight = fontHeight * 1.3
		end
	else
		text = CombatFeedbackText[event]
	end

	self.feedbackStartTime = GetTime()

	local font = self.HitIndicator:GetFont()
	self.HitIndicator:SetFont(font,fontHeight,"OUTLINE")
	self.HitIndicator:SetText(text)
	self.HitIndicator:SetTextColor(r, g, b)
	self.HitIndicator:SetAlpha(0)
	self.HitIndicator:Show()

	aUF.feedback[self.name] = true
	aUF:ScheduleRepeatingEvent("agUF_CombatSchedule",aUF.FeedbackUpdate, 0.05, aUF)
end

------------------
-- UNITPLAYER CLASS --
------------------

function aUF.classes.aUFunitPlayer.prototype:init(name,unit)
	aUF.classes.aUFunitPlayer.super.prototype.init(self,name,unit)
end

function aUF.classes.aUFunitPlayer.prototype:ClassRegisterEvents(register)
    if register then
        self:ObjectRegisterEvent("PLAYER_XP_UPDATE","UpdateXP")
        self:ObjectRegisterEvent("UPDATE_EXHAUSTION","UpdateXP")
        self:ObjectRegisterEvent("PLAYER_LEVEL_UP","UpdateXP")
        self:ObjectRegisterEvent("PLAYER_REGEN_ENABLED", "UpdateStatusIcon")
		self:ObjectRegisterEvent("PLAYER_REGEN_DISABLED", "UpdateStatusIcon")
        self:ObjectRegisterEvent("PLAYER_UPDATE_RESTING", "UpdateStatusIcon")
        if self.inCombatSchedule then
			self:CancelScheduledEvent(self.inCombatSchedule)
			self.inCombatSchedule = nil
		end    
    else
        local t = {"PLAYER_XP_UPDATE","UPDATE_EXHAUSTION","PLAYER_LEVEL_UP","PLAYER_REGEN_ENABLED","PLAYER_REGEN_DISABLED","PLAYER_UPDATE_RESTING"}
        for _,v in pairs(t) do 
            self:ObjectUnregisterEvent(v)
        end
    end
end

function aUF.classes.aUFunitPlayer.prototype:SetupFrame()
    self.class.super.prototype.SetupFrame(self)
    self.XPBar_Rest = CreateFrame("StatusBar",self.name.."_XPBar_Rest",self.frame)
	self.XPBar_Rest:SetMinMaxValues(0, 100)
	self.XPBar:SetParent(self.XPBar_Rest)
    
	self.XPBar:EnableMouse(true)
	self.XPBar:SetScript("OnEnter",function() self:XPOnEnter() end)
	self.XPBar:SetScript("OnLeave",function() GameTooltip:Hide() end)    
end

function aUF.classes.aUFunitPlayer.prototype:UpdateXP()
    if not self:IsHooked("ReputationWatchBar_Update") then
		self:SecureHook("ReputationWatchBar_Update", "UpdateXP")
	end
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
    if ( self.database.ShowXP == true ) then
        if repname then
            local color = aUF.RepColor[repreaction]
            repmax = repmax - repmin
            repvalue = repvalue - repmin
            repmin = 0
        
            if self.XPBar_Rest then
                self.XPBar_Rest:Hide()
            end
            
            self.XPBar:SetParent(self.frame)
            self.XPBar:Show()
            if repmax ~= 0 then
                self.XPBar:SetValue((repvalue/repmax)*100)
            else
                self.XPBar:SetValue(0)
            end
            self.XPBar:SetStatusBarColor(color.r, color.g, color.b)
            self.XPBar_BG:SetVertexColor(color.r, color.g, color.b, 0.25)
            self.XPBar_Rest:Hide()
        else
            local priorXP = self.XPBar:GetValue()
            local restXP = GetXPExhaustion()
            local currXP, nextXP = UnitXP(self.unit), UnitXPMax(self.unit)
                
            if nextXP ~= 0 then
                self.XPBar:SetValue((currXP/nextXP)*100)
            else
                self.XPBar:SetValue(0)
            end
            
            if restXP then
                self.XPBar_Rest:Show()
                self.Overlays:SetFrameLevel(self.XPBar:GetFrameLevel()+1)
                self.XPBar_Rest:SetFrameLevel(self.XPBar:GetFrameLevel()-1)
                self.XPBar:SetParent(self.XPBar_Rest)
                if nextXP ~= 0 then
                    self.XPBar_Rest:SetValue(((currXP+restXP)/nextXP)*100)
                else
                    self.XPBar_Rest:SetValue(0)
                end
            else
                self.XPBar:SetParent(self.frame)
                self.XPBar:Show()
                self.XPBar_Rest:Hide()
            end
            
            self.XPBar:SetStatusBarColor(0.8, 0, 0.7)
            if self.flags.BackgroundBarColor == true then
                self.XPBar_BG:SetVertexColor(0.8, 0, 0.7, 0.25)
            else
                self.XPBar_BG:SetVertexColor(0, 0, 0, 0.25)
            end
        end
    else
		self.XPBar:Hide()
		self.XPBar_BG:Hide()
        self.XPBar_Rest:Hide()    
    end
end

function aUF.classes.aUFunitPlayer.prototype:XPOnEnter()
	if (not this:IsVisible()) then return end
	
	GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
	GameTooltip:ClearLines()
	local _,eClass = UnitClass("player")
	
	local totalXP = UnitXPMax("player")
	local currentXP = UnitXP("player")
	local toLevelXP = totalXP - currentXP
	local restXP = GetXPExhaustion() or 0
	
	GameTooltip:AddLine(format("XP: %d/%d (%.1f%%)", currentXP, totalXP, currentXP/totalXP*100))
	if restXP == 0 then
		restXP = ""
	else
		restXP = format("|c0064e1ff+%d rested|r", restXP)
	end
	GameTooltip:AddLine(format("%d to level%s", toLevelXP, restXP))
	
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if (repname) then
		GameTooltip:AddLine(repname .. " (" .. getglobal("FACTION_STANDING_LABEL" .. repreaction) .. ")")
		GameTooltip:AddLine(format("%d/%d (%.1f%%)", repvalue, repmax, (repvalue-repmin)/(repmax-repmin) * 100));
    end
	GameTooltip:Show()
end

function aUF.classes.aUFunitPlayer.prototype:UpdateAll()
	self.class.super.prototype.UpdateAll(self)
	self:UpdateXP()
end

------------------
-- UNITPET CLASS --
------------------

function aUF.classes.aUFunitPet.prototype:init(name,unit)
	aUF.classes.aUFunitPet.super.prototype.init(self,name,unit)
end

function aUF.classes.aUFunitPet.prototype:ClassRegisterEvents(register)
    if register then
        self:ObjectRegisterEvent("UNIT_PET_EXPERIENCE","UpdateXP")
        self:ObjectRegisterEvent("UNIT_HAPPINESS","StatusBarsColor")
    else
        self:ObjectUnregisterEvent("UNIT_PET_EXPERIENCE")
        self:ObjectUnregisterEvent("UNIT_HAPPINESS")
    end
end

function aUF.classes.aUFunitPet.prototype:SetupFrame() 
    self.class.super.prototype.SetupFrame(self)
	self.XPBar:EnableMouse(true)
	self.XPBar:SetScript("OnEnter",function() self:XPOnEnter() end)
	self.XPBar:SetScript("OnLeave",function() GameTooltip:Hide() end)    
end

function aUF.classes.aUFunitPet.prototype:UpdateXP()
	local _,eClass = UnitClass("player")
	if eClass ~= "HUNTER" then
		return
	end
	if ( self.database.ShowXP == true and eClass == "HUNTER") then
        local priorXP = self.XPBar:GetValue()		
        local currXP, nextXP = GetPetExperience()
		
		if nextXP ~= 0 then
			self.XPBar:SetValue((currXP/nextXP)*100)
		else
			self.XPBar:SetValue(0)
		end
					
		self.XPBar:SetStatusBarColor(0.8, 0, 0.7)
		if self.flags.BackgroundBarColor == true then
			self.XPBar_BG:SetVertexColor(0.8, 0, 0.7, 0.25)
		else
			self.XPBar_BG:SetVertexColor(0, 0, 0, 0.25)
        end
	else
		self.XPBar:Hide()
		self.XPBar_BG:Hide()
		
	end
end

function aUF.classes.aUFunitPet.prototype:XPOnEnter()
	if (not this:IsVisible()) then return end
	
	GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
	GameTooltip:ClearLines()
	local _,eClass = UnitClass("player")
	
    if ( eClass == "HUNTER") then
		currXP, nextXP = GetPetExperience()
		local toLevelXP = nextXP - currXP
		
		GameTooltip:AddLine(format("Pet XP: %d/%d (%.1f%%)", currXP, nextXP, currXP/nextXP*100))
		GameTooltip:AddLine("" .. toLevelXP .. " to level")
		GameTooltip:AddLine()
		
		local happiness, damagePercentage, loyaltyRate = GetPetHappiness()
		happiness = ({"|cffff0000Unhappy|r", "|cffffff00Content|r", "|cff00ff00Happy|r"})[happiness]
		local loyalty = loyaltyRate > 0 and "Gaining" or "Losing"
		
		GameTooltip:AddLine("Pet is " .. happiness .. ", doing " .. damagePercentage .. "% damage")
		GameTooltip:AddLine(loyalty .. " loyalty")
	end
	GameTooltip:Show()
end

function aUF.classes.aUFunitPet.prototype:UpdateAll()
	self.class.super.prototype.UpdateAll(self)
	self:UpdateXP()
end

---------------------
-- TARGET SUBCLASS --
---------------------

function aUF.classes.aUFunitCombo.prototype:init(name,unit)
	aUF.classes.aUFunitCombo.super.prototype.init(self,name,unit)
	
end

function aUF.classes.aUFunitCombo.prototype:ClassRegisterEvents(register)
    if register then
        self:ObjectRegisterEvent("PLAYER_COMBO_POINTS","UpdateComboPoints")
    else
         self:ObjectUnregisterEvent("PLAYER_COMBO_POINTS")
    end
end

function aUF.classes.aUFunitCombo.prototype:SetupFrame()
    self.class.super.prototype.SetupFrame(self)
	for i=1,5 do
		self["Combo"..i] = self.Overlays:CreateTexture(nil,"OVERLAY")
		self["Combo"..i]:SetTexture("Interface\\AddOns\\ag_UnitFrames\\images\\combo.tga")
		self["Combo"..i]:SetHeight(10)
		self["Combo"..i]:SetWidth(10)
		self["Combo"..i]:Hide()
		if i > 1 then
			self["Combo"..i]:SetPoint("BOTTOMRIGHT",self["Combo"..i-1],"BOTTOMLEFT")
		else
			self["Combo"..i]:SetPoint("BOTTOMRIGHT",self.frame,"BOTTOMRIGHT",-2,-1);
		end
	end
end

function aUF.classes.aUFunitCombo.prototype:UpdateComboPoints()
	local points = GetComboPoints()
	
	if comboGFX == 1 then
		for i=0,4 do
			if points > i then
				self["Combo"..i+1]:Show()
			else
				self["Combo"..i+1]:Hide()
			end
		end
	else
		if points == 0 then
			points = ""
		end

		self.ComboText:SetText(points)
	end
end

function aUF.classes.aUFunitCombo.prototype:UpdateHighlight(entered)
    self.class.super.prototype.UpdateHighlight(self,entered)
    if UnitExists(self.unit) then
        self:UpdateAll(nil,true)
    end
    self:UpdateComboPoints()
end

---------------------
-- FOCUS SUBCLASS --
---------------------

function aUF.classes.aUFunitFocus.prototype:init(name,unit)
	aUF.classes.aUFunitFocus.super.prototype.init(self,name,unit)
end

function aUF.classes.aUFunitFocus.prototype:ClassRegisterEvents(register)
    if register then
        self:ObjectRegisterEvent("PLAYER_FOCUS_CHANGED")
    end
end

function aUF.classes.aUFunitFocus.prototype:PLAYER_FOCUS_CHANGED()
    if UnitExists(self.unit) then
        self:UpdateAll()
    end
end

-----------------------
-- ONUPDATE SUBCLASS --
-----------------------

function aUF.classes.aUFunitMetro.prototype:init(name,unit)
	aUF.classes.aUFunitMetro.super.prototype.init(self,name,unit)
    self.metro = true
end

function aUF.classes.aUFunitMetro.prototype:UpdateMetro(force)
	if force or self.frame:IsVisible() then
		self:UpdateHealth(true)
		self:UpdatePower(true)
		self:UpdateAuras(true)
		self:UpdateRaidTargetIcon(true)
		self:StatusBarsColor()
		self:UpdateTextStrings()
		if self.database.Portrait and self.flags.Portrait and self.unitName ~= UnitName(self.unit) then
			self:UpdatePortrait()
		end
	end
	self.unitName = UnitName(self.unit)
end

function aUF.classes.aUFunitMetro.prototype:Start()
	if not self.database.HideFrame == true then
		if not self.schedule then
			self.schedule = self:ScheduleRepeatingEvent(self.UpdateMetro, 0.5, self)
		end
		self:UpdateMetro()
	end
end

function aUF.classes.aUFunitMetro.prototype:Stop()
	if self.schedule and self:IsEventScheduled(self.schedule) then
		self:CancelScheduledEvent(self.schedule)
		self.schedule = nil
	end
	self:UpdateMetro()
end

function aUF.classes.aUFunitMetro.prototype:UpdateHighlight(entered)
    self.class.super.prototype.UpdateHighlight(self,entered)
	self:UpdateMetro(true)
end
--------------------------
-- UNITVANILLA SUBCLASS --
--------------------------

function aUF.classes.aUFunitVanilla.prototype:init(name,unit)
	aUF.classes.aUFunitVanilla.super.prototype.init(self,name,unit)
end
